from io import StringIO
from typing import Optional

from cirro.auth.device_code import DeviceCodeAuth

__all__ = [
    'get_auth_info_from_config',
    "DeviceCodeAuth"
]

from cirro.config import AppConfig


def get_auth_info_from_config(app_config: AppConfig, auth_io: Optional[StringIO] = None):
    user_config = app_config.user_config
    if not user_config or not user_config.auth_method:
        return DeviceCodeAuth(region=app_config.region,
                              client_id=app_config.client_id,
                              auth_endpoint=app_config.auth_endpoint,
                              auth_io=auth_io)

    auth_methods = [
        DeviceCodeAuth
    ]
    matched_auth_method = next((m for m in auth_methods if m.__name__ == user_config.auth_method), None)
    if not matched_auth_method:
        # Backwards compatibility
        if user_config.auth_method == 'ClientAuth':
            matched_auth_method = DeviceCodeAuth
        else:
            raise RuntimeError(f'{user_config.auth_method} not found, please re-run configuration')

    auth_config = user_config.auth_method_config

    if matched_auth_method == DeviceCodeAuth:
        return DeviceCodeAuth(region=app_config.region,
                              client_id=app_config.client_id,
                              auth_endpoint=app_config.auth_endpoint,
                              enable_cache=auth_config.get('enable_cache') == 'True',
                              auth_io=auth_io)



import threading
from datetime import datetime, timezone
from functools import partial
from typing import List, Dict

from cirro_api_client import CirroApiClient
from cirro_api_client.v1.api.file import generate_project_file_access_token
from cirro_api_client.v1.models import AWSCredentials, AccessType

from cirro.clients.s3 import S3Client
from cirro.file_utils import upload_directory, download_directory
from cirro.models.file import FileAccessContext, File, PathLike
from cirro.services.base import BaseService


class FileService(BaseService):
    """
    Service for interacting with files
    """
    enable_additional_checksum: bool
    transfer_retries: int
    _get_token_lock = threading.Lock()
    _read_token_cache: Dict[str, AWSCredentials] = {}

    def __init__(self, api_client, enable_additional_checksum, transfer_retries):
        """
        Instantiates the file service class
        """
        self._api_client = api_client
        self.enable_additional_checksum = enable_additional_checksum
        self.transfer_retries = transfer_retries

    def get_access_credentials(self, access_context: FileAccessContext) -> AWSCredentials:
        """
        Retrieves credentials to access files

        Args:
            access_context (cirro.models.file.FileAccessContext): File access context, use class methods to generate
        """
        access_request = access_context.file_access_request

        if access_request.access_type == AccessType.PROJECT_DOWNLOAD:
            return self._get_project_read_credentials(access_context)

        else:
            return generate_project_file_access_token.sync(
                client=self._api_client,
                project_id=access_context.project_id,
                body=access_context.file_access_request
            )

    def _get_project_read_credentials(self, access_context: FileAccessContext):
        """
        Retrieves credentials to read project data, this can be cached
        """
        access_request = access_context.file_access_request
        project_id = access_context.project_id
        with self._get_token_lock:
            cached_token = self._read_token_cache.get(project_id)

            if not cached_token or datetime.now(tz=timezone.utc) > cached_token.expiration:
                new_token = generate_project_file_access_token.sync(
                    client=self._api_client,
                    project_id=project_id,
                    body=access_request
                )

                self._read_token_cache[project_id] = new_token

        return self._read_token_cache[project_id]

    def get_file(self, file: File) -> bytes:
        """
        Gets the contents of a file

        Args:
            file (cirro.models.file.File):

        Returns:
            The raw bytes of the file
        """
        return self.get_file_from_path(file.access_context, file.relative_path)

    def get_file_from_path(self, access_context: FileAccessContext, file_path: str) -> bytes:
        """
        Gets the contents of a file by providing the path, used internally

        Args:
            access_context (cirro.models.file.FileAccessContext): File access context, use class methods to generate
            file_path (str): Relative path to file within dataset

        Returns:
            The raw bytes of the file
        """

        s3_client = S3Client(
            partial(self.get_access_credentials, access_context),
            self.enable_additional_checksum
        )

        full_path = f'{access_context.prefix}/{file_path}'.lstrip('/')

        return s3_client.get_file(access_context.bucket, full_path)

    def create_file(self, access_context: FileAccessContext, key: str,
                    contents: str, content_type: str) -> None:
        """
        Creates a file at the specified path

        Args:
            access_context (cirro.models.file.FileAccessContext): File access context, use class methods to generate
            key (str): Key of object to create
            contents (str): Content of object
            content_type (str):
        """

        s3_client = S3Client(
            partial(self.get_access_credentials, access_context),
            self.enable_additional_checksum
        )

        s3_client.create_object(
            key=key,
            contents=contents,
            content_type=content_type,
            bucket=access_context.bucket
        )

    def upload_files(self,
                     access_context: FileAccessContext,
                     directory: PathLike,
                     files: List[PathLike]) -> None:
        """
        Uploads a list of files from the specified directory

        Args:
            access_context (cirro.models.file.FileAccessContext): File access context, use class methods to generate
            directory (str|Path): Path to directory
            files (typing.List[str|Path]): List of paths to files within the directory
                must be the same type as directory.
        """

        s3_client = S3Client(
            partial(self.get_access_credentials, access_context),
            self.enable_additional_checksum
        )

        upload_directory(
            directory,
            files,
            s3_client,
            access_context.bucket,
            access_context.prefix,
            max_retries=self.transfer_retries
        )

    def download_files(self, access_context: FileAccessContext, directory: str, files: List[str]) -> None:
        """
        Download a list of files to the specified directory

        Args:
            access_context (cirro.models.file.FileAccessContext): File access context, use class methods to generate
            directory (str): download location
            files (List[str]): relative path of files to download
        """
        s3_client = S3Client(
            partial(self.get_access_credentials, access_context),
            self.enable_additional_checksum
        )

        download_directory(
            directory,
            files,
            s3_client,
            access_context.bucket,
            access_context.prefix
        )


class FileEnabledService(BaseService):
    """
    Not to be instantiated directly
    """
    _file_service: FileService

    def __init__(self, api_client: CirroApiClient, file_service: FileService):
        super().__init__(api_client)
        self._file_service = file_service



import json
import logging
import os
from pathlib import Path

import boto3
import pandas as pd

from cirro.models.s3_path import S3Path


def _write_json(dat, local_path: str, indent=4):
    with Path(local_path).open(mode="wt") as handle:
        return json.dump(dat, handle, indent=indent)


class PreprocessDataset:
    """
    Helper functions for performing preparatory tasks immediately before launching
    the analysis workflow for a dataset.
    """

    def __init__(self, s3_dataset: str, config_directory='config'):
        self.s3_dataset = s3_dataset
        self.files = self._read_csv(
            str(Path(config_directory, "files.csv")),
            required_columns=["sample", "file"]
        )

        self.samplesheet = self._read_csv(
            str(Path(config_directory, "samplesheet.csv")),
            required_columns=["sample"]
        )

        self.params = self._read_json(
            str(Path(config_directory, "params.json")),
        )

        self.metadata = self._read_json(
            str(Path(config_directory, "metadata.json")),
        )

        # Log to STDOUT
        log_formatter = logging.Formatter(
            '%(asctime)s %(levelname)-8s [PreprocessDataset] %(message)s'
        )
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(log_formatter)
        self.logger.addHandler(console_handler)

    @classmethod
    def from_running(cls):
        """
        Creates a dataset instance from the currently running dataset
          expected to be called from the headnode
        """
        dataset_path = os.getenv("PW_S3_DATASET")
        return cls(dataset_path)

    def log(self):
        """Print logging messages about the dataset."""

        self.logger.info(f"Storage location for dataset: {self.s3_dataset}")
        self.logger.info(f"Number of files in dataset: {self.files.shape[0]:,}")
        self.logger.info(f"Number of samples in dataset: {self.samplesheet.shape[0]:,}")

    def _read_csv(self, suffix: str, required_columns=None) -> pd.DataFrame:
        """Read a CSV from the dataset and check for any required columns."""
        if required_columns is None:
            required_columns = []

        df = pd.read_csv(f"{self.s3_dataset}/{suffix}")
        for col in required_columns:
            assert col in df.columns.values, f"Did not find expected columns {col} in {self.s3_dataset}/{suffix}"
        return df

    def _read_json(self, suffix: str):
        """Read a JSON from the dataset"""

        # Make the full S3 path
        path = f"{self.s3_dataset}/{suffix}"
        s3_path = S3Path(path)

        if s3_path.valid:
            s3 = boto3.client('s3')
            retr = s3.get_object(Bucket=s3_path.bucket, Key=s3_path.key)
            text = retr['Body'].read().decode()
        else:
            with Path(path).open() as handle:
                text = handle.read()

        # Parse JSON
        return json.loads(text)

    def add_param(self, name: str, value, overwrite=False, log=True):
        """Add a parameter to the dataset."""

        assert overwrite or name not in self.params, \
            f"Cannot add parameter {name}, already exists (and overwrite=False)"

        if log:
            self.logger.info(f"Adding parameter {name} = {value}")
        self.params[name] = value

        if log:
            self.logger.info("Saving parameters")
        _write_json(self.params, "nextflow.json")

    def remove_param(self, name: str, force=False):
        """Remove a parameter from the dataset."""

        assert force or name in self.params, \
            f"Cannot remove parameter {name}, does not exist (and force=False)"

        self.logger.info(f"Removing parameter {name}")
        if name in self.params:
            del self.params[name]

        self.logger.info("Saving parameters")
        _write_json(self.params, "nextflow.json")

    def update_compute(self, from_str, to_str, fp="nextflow-override.config"):
        """Replace all instances of a text string in the compute config file."""

        assert os.path.exists(fp), f"File does not exist: {fp}"
        with open(fp, 'r') as handle:
            compute = handle.read()
        n = len(compute.split(from_str)) - 1
        self.logger.info(f"Replacing {n:,} instances of {from_str} with {to_str} in {fp}")
        compute = compute.replace(from_str, to_str)
        with open(fp, 'wt') as handle:
            handle.write(compute)

    def wide_samplesheet(
            self,
            index=None,
            columns="read",
            values="file",
            column_prefix="fastq_"
    ):
        """Format a wide samplesheet with each read-pair on a row."""

        if index is None:
            index = ["sampleIndex", "sample", "lane"]
        self.logger.info("Formatting a wide samplesheet")
        self.logger.info("File table (long)")
        self.logger.info(self.files.head().to_csv(index=False))

        assert columns in self.files.columns.values, f"Column '{columns}' not found in file table"
        assert values in self.files.columns.values, f"Column '{values}' not found in file table"

        assert isinstance(index, list), f"index must be a list (not {type(index)})"

        # Get the list of columns from the inputs
        input_columns = self.files.columns.values

        # Format as a wide dataset
        # Note that all of the columns in `index` will be added if they are not already present
        wide_df = self.files.reindex(
            columns=index + [columns] + [values]
        ).pivot(
            index=index,
            columns=columns,
            values=values
        ).rename(
            columns=lambda i: f"{column_prefix}{int(i)}"
        ).reset_index(
        )

        # Remove any columns from the ouput which were added from `index`
        for cname in index:
            if cname not in input_columns:
                wide_df = wide_df.drop(columns=[cname])

        return wide_df



import logging
from typing import Tuple, Dict

from cirro.cli.interactive.utils import ask_yes_no, ask
from cirro.config import list_tenants, extract_base_url

logger = logging.getLogger()


def gather_auth_config() -> Tuple[str, str, Dict, bool]:
    try:
        tenant_options = list_tenants()
    except Exception:
        logger.exception('Failed to get tenant list, enter the URL manually')
        tenant_options = []

    base_url = ask(
        'autocomplete',
        'Enter the URL of the Cirro instance you\'d like to connect to (press TAB for options)',
        choices=[tenant['domain'] for tenant in tenant_options],
        meta_information={tenant['domain']: tenant['displayName'] for tenant in tenant_options}
    )
    # Fix user-provided base URL, if necessary
    base_url = extract_base_url(base_url)

    auth_method_config = {
        'enable_cache': ask_yes_no('Would you like to save your login? (do not use this on shared devices)')
    }

    enable_additional_checksum = ask(
        'select',
        'Upload / download file validation type (note: SHA-256 requires additional local compute)',
        choices=['MD5 (default)', 'SHA-256']
    ) == 'SHA-256'

    return 'ClientAuth', base_url, auth_method_config, enable_additional_checksum



from dataclasses import dataclass
from pathlib import PurePath, Path
from typing import Dict, Optional, TypeVar, NamedTuple

from cirro_api_client.v1.models import FileAccessRequest, AccessType, FileEntry

from cirro.models.s3_path import S3Path

PathLike = TypeVar('PathLike', str, Path)


class DirectoryStatistics(NamedTuple):
    size: float
    " Size in bytes"
    size_friendly: str
    " Size in user friendly format (e.g. 1.2 KB)"
    number_of_files: int
    " Number of files"


class FileAccessContext:
    """
    Context holder for accessing various files in Cirro and abstracting out their location.
    Prefer to use the class methods to instantiate.
    """
    def __init__(self,
                 file_access_request: FileAccessRequest,
                 project_id: str,
                 base_url: str):
        self.file_access_request = file_access_request
        self.base_url = base_url
        self.project_id = project_id
        self._s3_path = S3Path(base_url)

    @classmethod
    def download(cls, project_id: str, base_url: str, token_lifetime_override: int = None):
        return cls(
            file_access_request=FileAccessRequest(
                access_type=AccessType.PROJECT_DOWNLOAD,
                token_lifetime_hours=token_lifetime_override
            ),
            base_url=base_url,
            project_id=project_id
        )

    @classmethod
    def upload_dataset(cls, project_id: str, dataset_id: str, base_url: str, token_lifetime_override: int = None):
        return cls(
            file_access_request=FileAccessRequest(
                access_type=AccessType.DATASET_UPLOAD,
                dataset_id=dataset_id,
                token_lifetime_hours=token_lifetime_override
            ),
            base_url=f'{base_url}/data',
            project_id=project_id
        )

    @classmethod
    def upload_reference(cls, project_id: str, base_url: str):
        return cls(
            file_access_request=FileAccessRequest(
                access_type=AccessType.REFERENCE_UPLOAD
            ),
            base_url=base_url,
            project_id=project_id
        )

    @classmethod
    def upload_sample_sheet(cls, project_id: str, dataset_id: str, base_url: str):
        return cls(
            file_access_request=FileAccessRequest(
                access_type=AccessType.SAMPLESHEET_UPLOAD,
                dataset_id=dataset_id
            ),
            base_url=base_url,
            project_id=project_id
        )

    @property
    def bucket(self) -> str:
        """ S3 Bucket """
        return self._s3_path.bucket

    @property
    def prefix(self) -> str:
        """ S3 Prefix """
        return self._s3_path.key


@dataclass(frozen=True)
class File:
    relative_path: str
    size: int
    access_context: FileAccessContext
    metadata: Optional[Dict] = None

    @classmethod
    def from_file_entry(cls, file: FileEntry, project_id: str, domain: str = None):
        # Path is absolute rather than relative
        if 's3://' in file.path:
            parts = S3Path(file.path)
            domain = parts.base
            path = parts.key
        else:
            path = file.path

        return cls(
            relative_path=path,
            metadata=file.metadata.additional_properties,
            size=file.size,
            access_context=FileAccessContext.download(project_id=project_id, base_url=domain)
        )

    @property
    def absolute_path(self):
        return f'{self.access_context.base_url}/{self.relative_path.strip("/")}'

    @property
    def name(self):
        return PurePath(self.absolute_path).name

    def __repr__(self):
        return f'{self.__class__.__name__}(path={self.relative_path})'



from typing import List, Optional, Union

from cirro_api_client.v1.api.datasets import get_datasets, get_dataset, import_public_dataset, upload_dataset, \
    update_dataset, delete_dataset, get_dataset_manifest
from cirro_api_client.v1.models import ImportDataRequest, UploadDatasetRequest, UpdateDatasetRequest, Dataset, \
    DatasetDetail, CreateResponse, UploadDatasetCreateResponse

from cirro.models.file import FileAccessContext, File, PathLike
from cirro.services.base import get_all_records
from cirro.services.file import FileEnabledService


class DatasetService(FileEnabledService):
    """
    Service for interacting with the Dataset endpoints
    """

    def list(self, project_id: str, max_items: int = 10000) -> List[Dataset]:
        """List datasets

        Retrieves a list of datasets for a given project

        Args:
            project_id (str): ID of the Project
            max_items (int): Maximum number of records to get (default 10,000)
        """
        return get_all_records(
            records_getter=lambda page_args: get_datasets.sync(
                project_id=project_id,
                client=self._api_client,
                next_token=page_args.next_token,
                limit=page_args.limit
            ),
            max_items=max_items
        )

    def import_public(self, project_id: str, import_request: ImportDataRequest) -> CreateResponse:
        """
        Download data from public repositories

        Args:
            project_id (str): ID of the Project
            import_request (cirro_api_client.v1.models.ImportDataRequest):

        Returns:
            ID of the created dataset
        """
        return import_public_dataset.sync(project_id=project_id, client=self._api_client, body=import_request)

    def create(self, project_id: str, upload_request: UploadDatasetRequest) -> UploadDatasetCreateResponse:
        """
        Registers a dataset in Cirro, which can subsequently have files uploaded to it

        Args:
            project_id (str): ID of the Project
            upload_request (cirro_api_client.v1.models.UploadDatasetRequest):

        Returns:
            ID of the created dataset and the path to upload files

        ```python
        from cirro_api_client.v1.models import UploadDatasetRequest
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()
        request = UploadDatasetRequest(
            name="Name of new dataset",
            process_id="paired_dnaseq",
            expected_files=["read_1.fastq.gz", "read_2.fastq.gz"],
            description="Description of the dataset"
        )
        cirro.datasets.create("project-id", request)
        ```
        """
        return upload_dataset.sync(project_id=project_id, client=self._api_client, body=upload_request)

    def get(self, project_id: str, dataset_id: str) -> Optional[DatasetDetail]:
        """
        Gets detailed information about a dataset

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset

        Returns:
            The dataset, if found
        """
        return get_dataset.sync(project_id=project_id, dataset_id=dataset_id, client=self._api_client)

    def update(self, project_id: str, dataset_id: str, request: UpdateDatasetRequest) -> DatasetDetail:
        """
        Update info on a dataset (name, description, and/or tags)

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            request (cirro_api_client.v1.models.UpdateDatasetRequest):

        Returns:
            The updated dataset

        ```python
        from cirro_api_client.v1.models import UpdateDatasetRequest
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()
        request = UpdateDatasetRequest(
            name="Name of new dataset",
            process_id="paired_dnaseq",
            description="Description of the dataset"
        )
        cirro.datasets.update("project-id", "dataset-id", request)
        ```
        """
        return update_dataset.sync(project_id=project_id, dataset_id=dataset_id, body=request, client=self._api_client)

    def delete(self, project_id: str, dataset_id: str) -> None:
        """
        Delete a dataset

        After a dataset has been deleted, the files associated with that
        dataset are saved according to the project's retention time.

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
        """
        delete_dataset.sync_detailed(project_id=project_id, dataset_id=dataset_id, client=self._api_client)

    def get_file_listing(self, project_id: str, dataset_id: str, file_limit: int = 100000) -> List[File]:
        """
        Gets a listing of files, charts, and other assets available for the dataset

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            file_limit (int): Maximum number of files to get (default 100,000)
        """
        if file_limit < 1:
            raise ValueError("file_limit must be greater than 0")
        all_files = []
        file_offset = 0
        domain = None

        while len(all_files) < file_limit:
            manifest = get_dataset_manifest.sync(
                project_id=project_id,
                dataset_id=dataset_id,
                file_offset=file_offset,
                client=self._api_client
            )
            all_files.extend(manifest.files)
            file_offset += len(manifest.files)

            domain = manifest.domain
            if len(all_files) >= manifest.total_files or len(manifest.files) == 0:
                break

        files = [
            File.from_file_entry(
                f,
                project_id=project_id,
                domain=domain
            )
            for f in all_files
        ]
        return files

    def upload_files(self,
                     project_id: str,
                     dataset_id: str,
                     directory: PathLike,
                     files: List[PathLike]) -> None:
        """
        Uploads files to a given dataset from the specified directory.

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            directory (str|Path): Path to directory
            files (typing.List[str|Path]): List of paths to files within the directory,
                must be the same type as directory.
        """
        dataset = self.get(project_id, dataset_id)

        access_context = FileAccessContext.upload_dataset(
            project_id=project_id,
            dataset_id=dataset_id,
            base_url=dataset.s3
        )

        self._file_service.upload_files(
            access_context=access_context,
            directory=directory,
            files=files
        )

    def download_files(
        self,
        project_id: str,
        dataset_id: str,
        download_location: str,
        files: Union[List[File], List[str]] = None
    ) -> None:
        """
        Downloads files from a dataset

        The `files` argument is used to optionally specify a subset of files
        to be downloaded. By default, all files are downloaded.

        Args:
            project_id (str): ID of the Project
            dataset_id (str): ID of the Dataset
            download_location (str): Local destination for downloaded files
            files (typing.List[str]): Optional list of files to download
        """
        if files is None:
            files = self.get_file_listing(project_id, dataset_id)

        if len(files) == 0:
            return

        first_file = files[0]
        if isinstance(first_file, File):
            files = [file.relative_path for file in files]
            access_context = first_file.access_context
        else:
            dataset = self.get(project_id, dataset_id)
            access_context = FileAccessContext.download(project_id=project_id,
                                                        base_url=dataset.s3)

        self._file_service.download_files(access_context, download_location, files)



from pathlib import Path
from typing import List, Optional

from cirro_api_client.v1.api.processes import get_processes, get_process, get_process_parameters, \
    validate_file_requirements, archive_custom_process
from cirro_api_client.v1.models import ValidateFileRequirementsRequest, Executor, Process, ProcessDetail

from cirro.models.form_specification import ParameterSpecification
from cirro.services.base import BaseService


class ProcessService(BaseService):
    """
    Service for interacting with the Process endpoints
    """
    def list(self, process_type: Executor = None) -> List[Process]:
        """
        Retrieves a list of available processes

        Args:
            process_type (`cirro_api_client.v1.models.Executor`): Optional process type (INGEST, CROMWELL, or NEXTFLOW)
        """
        processes = get_processes.sync(client=self._api_client)
        return [p for p in processes if not process_type or process_type == p.executor]

    def get(self, process_id: str) -> ProcessDetail:
        """
        Retrieves detailed information on a process

        Args:
            process_id (str): Process ID
        """
        return get_process.sync(process_id=process_id, client=self._api_client)

    def archive(self, process_id: str):
        """
        Removes a custom process from the list of available processes.

        Error will be raised if the requested process does not exist. No value
        is returned, and no error raised if process exists and request is satisfied.

        Args:
            process_id (str): Process ID
        """
        archive_custom_process.sync_detailed(process_id=process_id, client=self._api_client)

    def find_by_name(self, name: str) -> Optional[ProcessDetail]:
        """
        Get a process by its display name

        Args:
            name (str): Process name
        """
        matched_process = next((p for p in self.list() if p.name == name), None)
        if not matched_process:
            return None

        return self.get(matched_process.id)

    def get_parameter_spec(self, process_id: str) -> ParameterSpecification:
        """
        Gets a specification used to describe the parameters used in the process

        Args:
            process_id (str): Process ID
        """
        form_spec = get_process_parameters.sync(process_id=process_id, client=self._api_client)
        return ParameterSpecification(form_spec)

    def check_dataset_files(self, files: List[str], process_id: str, directory: str):
        """
        Checks if the file mapping rules for a process are met by the list of files

        Error will be raised if the file mapping rules for the process are not met.
        No value is returned and no error is raised if the rules are satisfied.

        Args:
            process_id (str): ID for the process containing the file mapping rules
            directory: path to directory containing files
            files (List[str]): File names to check
        """
        # Parse sample sheet file if present
        sample_sheet = None
        sample_sheet_file = Path(directory, 'samplesheet.csv')
        if sample_sheet_file.exists():
            sample_sheet = sample_sheet_file.read_text()

        request = ValidateFileRequirementsRequest(
            file_names=files,
            sample_sheet=sample_sheet
        )
        requirements = validate_file_requirements.sync(process_id=process_id, body=request, client=self._api_client)

        # These will be sample sheet errors or no files errors
        if error_msg := requirements.error_msg:
            raise ValueError(error_msg)

        errors = [
            f'{entry.description}. {entry.error_msg}. We accept any of the following naming conventions: \n\t- ' +
            '\n\t- '.join([
                e.example_name
                for e in entry.allowed_patterns
            ])
            for entry in requirements.allowed_data_types
            if entry.error_msg is not None
        ]

        files_provided = ', '.join(files)

        if len(errors) != 0:
            raise ValueError(f"The files you have provided are: {files_provided} \n\n"
                             "They do not meet the dataset requirements. "
                             "The required file types are: \n" +
                             "\n".join(errors))



from typing import List

from cirro_api_client.v1.api.metadata import get_project_samples, get_project_schema, update_project_schema, \
    update_sample
from cirro_api_client.v1.models import FormSchema, SampleRequest, Sample

from cirro.services.base import BaseService, get_all_records


class MetadataService(BaseService):
    """
    Service for interacting with the Metadata endpoints
    """
    def get_project_samples(self, project_id: str, max_items: int = 10000) -> List[Sample]:
        """
        Retrieves a list of samples associated with a project along with their metadata

        Args:
            project_id (str): ID of the Project
            max_items (int): Maximum number of records to get (default 10,000)
        """
        return get_all_records(
            records_getter=lambda page_args: get_project_samples.sync(project_id=project_id,
                                                                      client=self._api_client,
                                                                      next_token=page_args.next_token,
                                                                      limit=page_args.limit),
            max_items=max_items
        )

    def get_project_schema(self, project_id: str) -> FormSchema:
        """
        Get project metadata schema

        Args:
            project_id (str): ID of the Project
        """
        return get_project_schema.sync(project_id=project_id, client=self._api_client)

    def update_project_schema(self, project_id: str, schema: FormSchema):
        """
        Update project metadata schema

        Args:
            project_id (str): ID of the Project
            schema (cirro_api_client.v1.models.FormSchema): Metadata schema
        """
        update_project_schema.sync_detailed(project_id=project_id, body=schema, client=self._api_client)

    def update_sample(self, project_id: str, sample_id: str, sample: SampleRequest) -> Sample:
        """
        Updates metadata information for sample

        Args:
            project_id (str): ID of the Project
            sample_id (str): ID of the sample
            sample (cirro_api_client.v1.models.SampleRequest): Metadata information for the sample
        """
        return update_sample.sync(
            project_id=project_id,
            sample_id=sample_id,
            body=sample,
            client=self._api_client
        )



from functools import cache
from time import sleep
from typing import List, Union

from cirro_api_client.v1.models import Project, UploadDatasetRequest, Dataset

from cirro.cirro_client import CirroApi
from cirro.file_utils import get_files_in_directory
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset
from cirro.sdk.dataset import DataPortalDataset, DataPortalDatasets
from cirro.sdk.exceptions import DataPortalAssetNotFound, DataPortalInputError
from cirro.sdk.helpers import parse_process_name_or_id
from cirro.sdk.process import DataPortalProcess
from cirro.sdk.reference import DataPortalReference, DataPortalReferences
from cirro.sdk.reference_type import DataPortalReferenceType, DataPortalReferenceTypes


class DataPortalProject(DataPortalAsset):
    """
    Projects in the Data Portal contain collections of Datasets.
    Users are granted permissions at the project-level, allowing them
    to view and/or modify all the datasets in that collection.
    """
    def __init__(self, proj: Project, client: CirroApi):
        """
        Instantiate with helper method

        ```python
        from cirro import DataPortal()
        portal = DataPortal()
        project = portal.get_project_by_name("Project Name")
        ```

        """
        self._data = proj
        self._client = client

    @property
    def id(self) -> str:
        """
        Unique identifier
        """
        return self._data.id

    @property
    def name(self) -> str:
        """
        Readable name
        """
        return self._data.name

    @property
    def description(self) -> str:
        """
        Longer description of the project
        """
        return self._data.description

    def __str__(self):
        """Control how the Project is rendered as a string."""

        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'id', 'description']
        ])

    @cache
    def _get_datasets(self) -> List[Dataset]:
        return self._client.datasets.list(self.id)

    def list_datasets(self, force_refresh=False) -> DataPortalDatasets:
        """List all the datasets available in the project."""
        if force_refresh:
            self._get_datasets.cache_clear()

        return DataPortalDatasets(
            [
                DataPortalDataset(d, self._client)
                for d in self._get_datasets()
            ]
        )

    def get_dataset_by_name(self, name: str, force_refresh=False) -> DataPortalDataset:
        """Return the dataset with the specified name."""
        if force_refresh:
            self._get_datasets.cache_clear()

        dataset = next((d for d in self._get_datasets() if d.name == name), None)
        if dataset is None:
            raise DataPortalAssetNotFound(f'Dataset with name {name} not found')
        return self.get_dataset_by_id(dataset.id)

    def get_dataset_by_id(self, _id: str = None) -> DataPortalDataset:
        """Return the dataset with the specified id."""

        dataset = self._client.datasets.get(project_id=self.id, dataset_id=_id)
        if dataset is None:
            raise DataPortalAssetNotFound(f'Dataset with ID {_id} not found')
        return DataPortalDataset(dataset, self._client)

    def list_references(self, reference_type: str = None) -> DataPortalReferences:
        """
        List the references available in a project.
        Optionally filter to references of a particular type (identified by name)
        """

        # Get the complete list of references which are available
        reference_types = DataPortalReferenceTypes(
            [
                DataPortalReferenceType(ref)
                for ref in self._client.references.get_types()
            ]
        )

        # If a particular name was specified
        if reference_type is not None:
            reference_types = reference_types.filter_by_pattern(reference_type)
            if len(reference_types) == 0:
                msg = f"Could not find any reference types with the name {reference_type}"
                raise DataPortalAssetNotFound(msg)

        return DataPortalReferences(
            [
                DataPortalReference(ref, project_id=self.id, client=self._client)
                for ref in self._client.references.get_for_project(
                    self.id
                )
                if reference_type is None or ref.type == reference_type
            ]
        )

    def get_reference_by_name(self, name: str = None, ref_type: str = None) -> DataPortalReference:
        """Return the reference of a particular type with the specified name."""

        if name is None:
            raise DataPortalInputError("Must specify the reference name")

        return self.list_references(ref_type).get_by_name(name)

    def upload_dataset(
        self,
        name: str = None,
        description='',
        process: Union[DataPortalProcess, str] = None,
        upload_folder: str = None,
        files: list = None
    ):
        """
        Upload a set of files to the Data Portal, creating a new dataset.

        If the files parameter is not provided, it will upload all files in the upload folder

        Args:
            name (str): Name of newly created dataset
            description (str): Description of newly created dataset
            process (str | DataPortalProcess): Process to run may be referenced by name, ID, or object
            upload_folder (str): Folder containing files to upload
            files (List[str]): Optional subset of files to upload from the folder
        """

        if name is None:
            raise DataPortalInputError("Must provide name for new dataset")
        if process is None:
            raise DataPortalInputError("Must provide the process which is used for ingest")
        if upload_folder is None:
            raise DataPortalInputError("Must provide upload_folder -- folder containing files to upload")

        # Parse the process provided by the user
        process = parse_process_name_or_id(process, self._client)

        # If no files were provided
        if files is None:
            # Get the list of files in the upload folder
            files = get_files_in_directory(upload_folder)

        if files is None or len(files) == 0:
            raise RuntimeWarning("No files to upload, exiting")

        # Make sure that the files match the expected pattern
        self._client.processes.check_dataset_files(files, process.id, upload_folder)

        # Create the ingest process request
        dataset_create_request = UploadDatasetRequest(
            process_id=process.id,
            name=name,
            description=description,
            expected_files=files
        )

        # Get the response
        create_response = self._client.datasets.create(project_id=self.id,
                                                       upload_request=dataset_create_request)

        # Upload the files
        self._client.datasets.upload_files(
            project_id=self.id,
            dataset_id=create_response.id,
            directory=upload_folder,
            files=files
        )

        # Return the dataset which was created, which might take a second to update
        max_attempts = 5
        for attempt in range(max_attempts):
            try:
                return self.get_dataset_by_id(create_response.id)
            except DataPortalAssetNotFound as e:
                if attempt == max_attempts - 1:
                    raise e
                else:
                    sleep(2)

    def samples(self, max_items: int = 10000):
        """
        Retrieves a list of samples associated with a project along with their metadata

        Args:
            max_items (int): Maximum number of records to get (default 10,000)
        """
        return self._client.metadata.get_project_samples(self.id, max_items)


class DataPortalProjects(DataPortalAssets[DataPortalProject]):
    """Collection of DataPortalProject objects"""
    asset_name = "project"



from abc import ABC
from typing import TypeVar, Callable, Generic, Optional, List

from attr import define
from cirro_api_client import CirroApiClient

D = TypeVar('D')


@define
class PageResp(Generic[D]):
    data: D
    next_token: str


@define
class PageArgs:
    next_token: str
    limit: int


T = TypeVar('T', bound=PageResp)


def get_all_records(records_getter: Callable[[PageArgs], Optional[PageResp[D]]],
                    batch_size=5000, max_items=None) -> List[D]:
    next_token = None
    items = []

    while True:
        resp = records_getter(PageArgs(next_token=next_token, limit=batch_size))
        if not resp:
            return items

        items.extend(resp.data)

        next_token = resp.next_token
        if not next_token:
            return items

        if max_items and len(items) >= max_items:
            return items


@define
class BaseService(ABC):
    """
    Not to be instantiated directly
    """
    _api_client: CirroApiClient



import os
import random
import time
from pathlib import Path, PurePath
from typing import List, Union

from boto3.exceptions import S3UploadFailedError
from botocore.exceptions import ConnectionError

from cirro.clients import S3Client
from cirro.models.file import DirectoryStatistics, File, PathLike

if os.name == 'nt':
    import win32api
    import win32con


def filter_files_by_pattern(files: Union[List[File], List[str]], pattern: str) -> Union[List[File], List[str]]:
    """
    Filters a list of files by a glob pattern

    Args:
        files (Union[List[File], List[str]]): List of Files or file paths
        pattern (str): Glob pattern (i.e., *.fastq)

    Returns:
        The filtered list of files
    """
    def matches_glob(file: Union[File, str]):
        return PurePath(file if isinstance(file, str) else file.relative_path).match(pattern)

    return [
        file for file in files
        if matches_glob(file)
    ]


def _is_hidden_file(file_path: Path):
    # Remove hidden files from listing, desktop.ini .DS_Store, etc.
    if os.name == 'nt':
        attributes = win32api.GetFileAttributes(str(file_path))
        return attributes & (win32con.FILE_ATTRIBUTE_HIDDEN | win32con.FILE_ATTRIBUTE_SYSTEM)
    else:
        return file_path.name.startswith('.')


def get_files_in_directory(
    directory: Union[str, Path],
    include_hidden=False
) -> List[str]:
    """
    Returns a list of strings containing the relative path of
    each file within the indicated directory.

    Args:
        directory (Union[str, Path]): The path to the directory
        include_hidden (bool): include hidden files in the returned list

    Returns:
        List of files in the directory
    """
    path = Path(directory).expanduser()
    path_posix = str(path.as_posix())

    paths = []

    for file_path in path.rglob("*"):
        if file_path.is_dir():
            continue

        if not include_hidden and _is_hidden_file(file_path):
            continue

        if not file_path.exists():
            continue

        str_file_path = str(file_path.as_posix())
        str_file_path = str_file_path.replace(f'{path_posix}/', "")
        paths.append(str_file_path)

    paths.sort()
    return paths


def _bytes_to_human_readable(num_bytes: int) -> str:
    for unit in ['B', 'KB', 'MB', 'GB', 'TB', 'PB']:
        if num_bytes < 1000.0 or unit == 'PB':
            break
        num_bytes /= 1000.0
    return f"{num_bytes:,.2f} {unit}"


def get_files_stats(files: List[PathLike]) -> DirectoryStatistics:
    """
    Returns information about the list of files provided, such as the total size and number of files.
    """
    sizes = [f.stat().st_size for f in files]
    total_size = sum(sizes)
    return DirectoryStatistics(
        size_friendly=_bytes_to_human_readable(total_size),
        size=total_size,
        number_of_files=len(sizes)
    )


def upload_directory(directory: PathLike,
                     files: List[PathLike],
                     s3_client: S3Client,
                     bucket: str,
                     prefix: str,
                     max_retries=10):
    """
    @private

    Uploads a list of files from the specified directory
    Args:
        directory (str|Path): Path to directory
        files (typing.List[str|Path]): List of paths to files within the directory
            must be the same type as directory.
        s3_client (cirro.clients.S3Client): S3 client
        bucket (str): S3 bucket
        prefix (str): S3 prefix
        max_retries (int): Number of retries
    """
    # Ensure all files are of the same type as the directory
    if not all(isinstance(file, type(directory)) for file in files):
        raise ValueError("All files must be of the same type as the directory (str or Path)")

    for file in files:
        if isinstance(file, str):
            file_path = Path(directory, file)
        else:
            file_path = file

        file_relative = file_path.relative_to(directory).as_posix()
        key = f'{prefix}/{file_relative}'
        success = False

        # Retry up to max_retries times
        for retry in range(max_retries):

            # Try the upload
            try:
                s3_client.upload_file(
                    file_path=file_path,
                    bucket=bucket,
                    key=key
                )

                success = True

            # Catch the upload error
            except (S3UploadFailedError, ConnectionError) as e:
                delay = random.uniform(0, 60) + retry * 60
                # Report the error
                print(f"Encountered error:\n{str(e)}\n"
                      f"Retrying in {delay:.0f} seconds ({max_retries - (retry + 1)} attempts remaining)")
                time.sleep(delay)

            if success:
                break


def download_directory(directory: str, files: List[str], s3_client: S3Client, bucket: str, prefix: str):
    """
    @private
    """
    for file in files:
        key = f'{prefix}/{file}'.lstrip('/')
        local_path = Path(directory, file)
        local_path.parent.mkdir(parents=True, exist_ok=True)

        s3_client.download_file(local_path=local_path,
                                bucket=bucket,
                                key=key)



import gzip
from io import BytesIO, StringIO
from typing import List

import pandas as pd

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import anndata

from cirro.cirro_client import CirroApi
from cirro.models.file import File
from cirro.sdk.asset import DataPortalAssets, DataPortalAsset
from cirro.sdk.exceptions import DataPortalInputError
from cirro.utils import convert_size


class DataPortalFile(DataPortalAsset):
    """
    Datasets are made up of a collection of File objects in the Data Portal.
    """

    def __init__(self, file: File, client: CirroApi):
        """
        Instantiate by listing files from a dataset.

        ```python
        from cirro import DataPortal()
        portal = DataPortal()
        dataset = portal.get_dataset(
            project="id-or-name-of-project",
            dataset="id-or-name-of-dataset"
        )
        files = dataset.list_files()
        ```
        """
        # Attach the file object
        self._file = file
        self._client = client

    # Note that the 'name' and 'id' attributes are set to the relative path
    # The purpose of this is to support the DataPortalAssets class functions
    @property
    def id(self) -> str:
        """Relative path of file within the dataset"""
        return self._file.relative_path

    @property
    def name(self) -> str:
        """Relative path of file within the dataset"""
        return self._file.relative_path

    @property
    def file_name(self) -> str:
        """Name of file, excluding the full folder path within the dataset"""
        return self._file.name

    @property
    def relative_path(self) -> str:
        """Relative path of file within the dataset"""
        return self._file.relative_path

    @property
    def absolute_path(self) -> str:
        """Fully URI to file object in AWS S3"""
        return self._file.absolute_path

    @property
    def metadata(self) -> dict:
        """File metadata"""
        return self._file.metadata

    @property
    def size_bytes(self) -> int:
        """File size (in bytes)"""
        return self._file.size

    @property
    def size(self) -> str:
        """File size converted to human-readable (e.g., 4.50 GB)"""
        return convert_size(self._file.size)

    def __str__(self):
        return f"{self.relative_path} ({self.size})"

    def _get(self) -> bytes:
        """Internal method to call client.file.get_file"""

        return self._client.file.get_file(self._file)

    def read_csv(self, compression='infer', encoding='utf-8', **kwargs) -> pd.DataFrame:
        """
        Parse the file as a Pandas DataFrame.

        The default field separator is a comma (for CSV), use sep='\\t' for TSV.

        File compression is inferred from the extension, but can be set
        explicitly with the compression= flag.

        All other keyword arguments are passed to pandas.read_csv
        https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
        """

        if compression == 'infer':
            # If the file appears to be compressed
            if self.relative_path.endswith('.gz'):
                compression = dict(method='gzip')
            elif self.relative_path.endswith('.bz2'):
                compression = dict(method='bz2')
            elif self.relative_path.endswith('.xz'):
                compression = dict(method='zstd')
            elif self.relative_path.endswith('.zst'):
                compression = dict(method='zstd')
            else:
                compression = None

        if compression is not None:
            handle = BytesIO(self._get())
        else:
            handle = StringIO(self._get().decode(encoding))

        df = pd.read_csv(
            handle,
            compression=compression,
            encoding=encoding,
            **kwargs
        )
        handle.close()
        return df

    def read_h5ad(self) -> 'anndata.AnnData':
        """Read an AnnData object from a file."""
        # Import the anndata library, and raise an error if it is not available
        try:
            import anndata as ad # noqa
        except ImportError:
            raise ImportError("The anndata library is required to read AnnData files. "
                              "Please install it using 'pip install anndata'.")

        # Download the file to a temporary file handle and parse the contents
        with BytesIO(self._get()) as handle:
            return ad.read_h5ad(handle)

    def readlines(self, encoding='utf-8', compression=None) -> List[str]:
        """Read the file contents as a list of lines."""

        return self.read(
            encoding=encoding,
            compression=compression
        ).splitlines()

    def read(self, encoding='utf-8', compression=None) -> str:
        """Read the file contents as text."""

        # Get the raw file contents
        cont = self._get()

        # If the file is uncompressed
        if compression is None:
            return cont.decode(encoding)
        # If the file is compressed
        else:

            # Only gzip-compression is supported currently
            if compression != "gzip":
                raise DataPortalInputError("compression may be 'gzip' or None")

            with gzip.open(
                BytesIO(
                    cont
                ),
                'rt',
                encoding=encoding
            ) as handle:
                return handle.read()

    def download(self, download_location: str = None):
        """Download the file to a local directory."""

        if download_location is None:
            raise DataPortalInputError("Must provide download location")

        self._client.file.download_files(
            self._file.access_context,
            download_location,
            [self.relative_path]
        )


class DataPortalFiles(DataPortalAssets[DataPortalFile]):
    """Collection of DataPortalFile objects."""
    asset_name = "file"

    def download(self, download_location: str = None) -> None:
        """Download the collection of files to a local directory."""

        for f in self:
            f.download(download_location)



from cirro_api_client.v1.models import Executor

from cirro.cirro_client import CirroApi
from cirro.sdk.dataset import DataPortalDataset
from cirro.sdk.exceptions import DataPortalAssetNotFound
from cirro.sdk.process import DataPortalProcess, DataPortalProcesses
from cirro.sdk.project import DataPortalProject, DataPortalProjects
from cirro.sdk.reference_type import DataPortalReferenceType, DataPortalReferenceTypes


class DataPortal:
    """
    Helper functions for exploring the Projects, Datasets, Samples, and Files
    available in the Data Portal.
    """

    def __init__(self, base_url: str = None, client: CirroApi = None):
        """
        Set up the DataPortal object, establishing an authenticated connection.

        Args:
            base_url (str): Optional base URL of the Cirro instance
             (if not provided, it uses the `CIRRO_BASE_URL` environment variable, or the config file)
            client (`cirro.cirro_client.CirroApi`): Optional pre-configured client

        Example:
        ```python
        from cirro import DataPortal

        Portal = DataPortal(base_url="app.cirro.bio")
        portal.list_projects()
        ```
        """

        if client is not None:
            self._client = client

        # Set up default client if not provided
        else:
            self._client = CirroApi(base_url=base_url)

    def list_projects(self) -> DataPortalProjects:
        """List all the projects available in the Data Portal."""

        return DataPortalProjects(
            [
                DataPortalProject(proj, self._client)
                for proj in self._client.projects.list()
            ]
        )

    def get_project_by_name(self, name: str = None) -> DataPortalProject:
        """Return the project with the specified name."""

        return self.list_projects().get_by_name(name)

    def get_project_by_id(self, _id: str = None) -> DataPortalProject:
        """Return the project with the specified id."""

        return self.list_projects().get_by_id(_id)

    def get_dataset(self, project: str = None, dataset: str = None) -> DataPortalDataset:
        """
        Return a dataset identified by ID or name.

        Args:
            project (str): ID or name of project
            dataset (str): ID or name of dataset

        Returns:
            `cirro.sdk.dataset.DataPortalDataset`

            ```python
            from cirro import DataPortal()
            portal = DataPortal()
            dataset = portal.get_dataset(
                project="id-or-name-of-project",
                dataset="id-or-name-of-dataset"
            )
            ```
        """
        try:
            project: DataPortalProject = self.get_project_by_id(project)
        except DataPortalAssetNotFound:
            project: DataPortalProject = self.get_project_by_name(project)

        try:
            return project.get_dataset_by_id(dataset)
        except DataPortalAssetNotFound:
            return project.get_dataset_by_name(dataset)

    def list_processes(self, ingest=False) -> DataPortalProcesses:
        """
        List all the processes available in the Data Portal.
        By default, only list non-ingest processes (those which can be run on existing datasets).
        To list the processes which can be used to upload datasets, use `ingest = True`.

        Args:
            ingest (bool): If True, only list those processes which can be used to ingest datasets directly
        """

        return DataPortalProcesses(
            [
                DataPortalProcess(p, self._client)
                for p in self._client.processes.list()
                if not ingest or p.executor == Executor.INGEST
            ]
        )

    def get_process_by_name(self, name: str, ingest=False) -> DataPortalProcess:
        """
        Return the process with the specified name.

        Args:
            name (str): Name of process
        """

        return self.list_processes(ingest=ingest).get_by_name(name)

    def get_process_by_id(self, id: str, ingest=False) -> DataPortalProcess:
        """
        Return the process with the specified id

        Args:
            id (str): ID of process
        """

        return self.list_processes(ingest=ingest).get_by_id(id)

    def list_reference_types(self) -> DataPortalReferenceTypes:
        """
        Return the list of all available reference types
        """

        return DataPortalReferenceTypes(
            [
                DataPortalReferenceType(ref)
                for ref in self._client.references.get_types()
            ]
        )



from cirro_api_client.v1.models import ReferenceType

from cirro.sdk.asset import DataPortalAssets, DataPortalAsset


class DataPortalReferenceType(DataPortalAsset):
    """
    Reference data is organized by project, categorized by type.
    """

    def __init__(self, ref_type: ReferenceType):
        self._data = ref_type

    @property
    def name(self):
        """Name of reference type"""
        return self._data.name

    @property
    def description(self):
        """Description of reference type"""
        return self._data.description

    @property
    def directory(self):
        return self._data.directory

    @property
    def validation(self):
        return self._data.validation

    def __str__(self):
        return '\n'.join([
            f"{i.title()}: {self.__getattribute__(i)}"
            for i in ['name', 'description']
        ])


class DataPortalReferenceTypes(DataPortalAssets[DataPortalReferenceType]):
    """Collection of DataPortalReferenceType objects."""
    asset_name = "reference type"

    def get_by_id(self, _id: str) -> DataPortalReferenceType:
        raise NotImplementedError("Filtering by ID is not supported, use get_by_name")



import json
import math
from datetime import timezone, datetime
from typing import Optional, Union

from cirro_api_client.v1.models import AWSCredentials


def parse_json_date(json_date: str) -> Optional[datetime]:
    if not json_date:
        return None
    local_zone = datetime.now(timezone.utc).astimezone().tzinfo
    try:
        parsed = datetime.strptime(json_date, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        # AWS Format
        parsed = datetime.strptime(json_date, "%Y-%m-%dT%H:%M:%S+00:00")

    return parsed\
        .replace(tzinfo=timezone.utc)\
        .astimezone(local_zone)


def safe_load_json(json_str: Optional[str]):
    if json_str is None:
        return {}

    return json.loads(json_str)


def format_date(date: Union[str, datetime]) -> str:
    if isinstance(date, str):
        date = parse_json_date(date)
    return date.strftime('%m/%d/%Y, %H:%M:%S')


def print_credentials(creds: AWSCredentials):
    print(f'AWS_ACCESS_KEY_ID={creds.access_key_id}')
    print(f'AWS_SECRET_ACCESS_KEY={creds.secret_access_key}')
    print(f'AWS_SESSION_TOKEN={creds.session_token}')
    if creds.expiration:
        print()
        print(f'These credentials expire at {format_date(creds.expiration)}')


def convert_size(size: int):
    if size == 0:
        return '0B'
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size, 1024)))
    p = math.pow(1024, i)
    s = round(size/p, 2)
    return '%.2f %s' % (s, size_name[i])



from typing import List

from cirro_api_client.v1.api.billing import get_billing_accounts, update_billing_account
from cirro_api_client.v1.models import BillingAccountRequest, BillingAccount

from cirro.services.base import BaseService


class BillingService(BaseService):
    """
    Service for interacting with the Billing endpoints
    """
    def list(self) -> List[BillingAccount]:
        """
        Gets a list of billing accounts the current user has access to
        """
        return get_billing_accounts.sync(client=self._api_client)

    def update(self, billing_account_id: str, request: BillingAccountRequest):
        """
        Updates a billing account

        Args:
            billing_account_id (str): Billing account ID
            request (cirro_api_client.v1.models.BillingAccountRequest):

        ```python
        from cirro_api_client.v1.models import BillingAccountRequest
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()
        request = BillingAccountRequest(
            name="New billing account name",
            primary_budget_number="new-budget-number",
            owner="New Owner"
        )
        cirro.billing.update("billing-account-id", request)
        ```
        """
        update_billing_account.sync_detailed(
            billing_account_id=billing_account_id,
            body=request,
            client=self._api_client
        )



from typing import List

from cirro_api_client.v1.api.users import invite_user, list_users, get_user
from cirro_api_client.v1.models import InviteUserRequest, User, UserDetail

from cirro.services.base import BaseService, get_all_records


class UserService(BaseService):
    def list(self, max_items: int = 10000) -> List[User]:
        """
        List users in the system
        """
        return get_all_records(
            records_getter=lambda page_args: list_users.sync(
                client=self._api_client,
                next_token=page_args.next_token,
                limit=page_args.limit
            ),
            max_items=max_items
        )

    def get(self, username: str) -> UserDetail:
        """
        Get user details by username, including what projects they are assigned to
        """
        return get_user.sync(username=username, client=self._api_client)

    def invite_user(self, name: str, organization: str, email: str):
        """
        Invite a user to the system.
        If the user already exists, it will return a message that the user already exists.

        Args:
            name (str): Name of the user
            organization (str): Organization of the user
            email (str): Email (username) of the user
        """
        request = InviteUserRequest(
            name=name,
            organization=organization,
            email=email,
        )
        response = invite_user.sync(client=self._api_client, body=request)
        return response.message



from typing import List, Optional

from cirro_api_client.v1.api.references import get_reference_types, get_references_for_project
from cirro_api_client.v1.models import ReferenceType, Reference

from cirro.services.base import BaseService


class ReferenceService(BaseService):
    """
    Service for interacting with the References endpoints
    """
    def get_types(self) -> Optional[List[ReferenceType]]:
        """
        List available reference types
        """
        return get_reference_types.sync(client=self._api_client)

    def get_for_project(self, project_id: str) -> Optional[List[Reference]]:
        """
        List available references for a given project
        """
        return get_references_for_project.sync(project_id=project_id, client=self._api_client)



from typing import List, Optional

from cirro_api_client.v1.api.projects import get_projects, create_project, get_project, set_user_project_role, \
    update_project, update_project_tags, get_project_users, archive_project, unarchive_project
from cirro_api_client.v1.models import ProjectRequest, SetUserProjectRoleRequest, Tag, ProjectRole, ProjectDetail, \
    CreateResponse, ProjectUser

from cirro.services.base import BaseService


class ProjectService(BaseService):
    """
    Service for interacting with the Project endpoints
    """
    def list(self):
        """
        Retrieve a list of projects
        """
        return get_projects.sync(client=self._api_client)

    def get(self, project_id: str) -> ProjectDetail:
        """
        Get detailed project information

        Args:
            project_id (str): Project ID
        """
        return get_project.sync(project_id=project_id, client=self._api_client)

    def create(self, request: ProjectRequest) -> CreateResponse:
        """
        Create a project

        Args:
            request (`cirro_api_client.v1.models.ProjectRequest`): Detailed information about the project to create

        Examples:
        ```python
        from cirro_api_client.v1.models import ProjectRequest, ProjectSettings, Contact, BudgetPeriod, CloudAccount, \
            CloudAccountType
        from cirro.cirro_client import CirroApi

        cirro = CirroApi()

        # Bring your own account projects are hosted by the user
        # You must provide the account details and VPC information
        # Please view the ProjectSettings model for more information on the attributes required
        byoa_project = ProjectRequest(
            name="New Project Name",
            description="Description of new project",
            billing_account_id="billing-account-id",
            settings=ProjectSettings(
                budget_period=BudgetPeriod.MONTHLY,
                budget_amount=1000,
                max_spot_vcpu=300,
                service_connections=[],
                retention_policy_days=7,
                vpc_id="vpc-000000000000",
                batch_subnets=["subnet-000000000000", "subnet-000000000001"],
                sagemaker_subnets=["subnet-000000000000", "subnet-000000000001"],
                kms_arn="arn:aws:kms:us-west-2:000000000000:key/00000000-0000-0000-0000-000000000000"
            ),
            account=CloudAccount(
                account_id="<AWS_ACCOUNT_ID>",
                region_name="us-west-2",
                account_name="Cirro Lab Project",
                account_type=CloudAccountType.BYOA
            ),
            contacts=[
                Contact(
                    name="Contact Name",
                    organization="Contact Organization",
                    email="contact@email.com",
                    phone="987-654-3210"
                )
            ]
        )

        # Hosted projects are managed by Cirro
        hosted_project = ProjectRequest(
            name="New Project Name",
            description="Description of new project",
            billing_account_id="billing-account-id",
            settings=ProjectSettings(
                budget_period=BudgetPeriod.MONTHLY,
                budget_amount=1000,
                max_spot_vcpu=300,
                service_connections=[],
                retention_policy_days=7
            ),
            contacts=[
                Contact(
                    name="Contact Name",
                    organization="Contact Organization",
                    email="contact@email.com",
                    phone="987-654-3210"
                )
            ]
        )

        cirro.projects.create(byoa_project)
        ```
        """
        return create_project.sync(client=self._api_client, body=request)

    def update(self, project_id: str, request: ProjectRequest) -> ProjectDetail:
        """
        Updates a project

        Args:
            project_id (str): ID of project to update
            request (`cirro_api_client.v1.models.ProjectRequest`): New details for the project
        """
        return update_project.sync(project_id=project_id, body=request, client=self._api_client)

    def update_tags(self, project_id: str, tags: List[Tag]):
        """
        Sets tags on a project

        Args:
            project_id (str): Project ID
            tags (List[Tag]): List of tags to apply
        """
        update_project_tags.sync_detailed(project_id=project_id, body=tags, client=self._api_client)

    def get_users(self, project_id: str) -> Optional[List[ProjectUser]]:
        """
        Gets users who have access to the project

        Args:
            project_id (str): Project ID
        """
        return get_project_users.sync(project_id=project_id, client=self._api_client)

    def set_user_role(self, project_id: str, username: str, role: ProjectRole, suppress_notification=False):
        """
        Sets a user's role within a project.

        Set to `ProjectRole.NONE` if removing the user from a project.

        Args:
            project_id (str): Project ID
            username (str): Username
            role (`cirro_api_client.v1.models.ProjectRole`): New role to apply
            suppress_notification (bool): Suppress email notification
        """
        request_body = SetUserProjectRoleRequest(
            username=username,
            role=role,
            suppress_notification=suppress_notification
        )
        set_user_project_role.sync_detailed(project_id=project_id, body=request_body, client=self._api_client)

    def archive(self, project_id: str):
        """
        Sets the project status to archived (hidden from active projects)

        Args:
            project_id (str): Project ID
        """
        archive_project.sync_detailed(project_id=project_id, client=self._api_client)

    def unarchive(self, project_id: str):
        """
        Sets the project status to active

        Args:
            project_id (str): Project ID
        """
        unarchive_project.sync_detailed(project_id=project_id, client=self._api_client)



import threading
from pathlib import Path
from typing import Callable

from boto3 import Session
from botocore.config import Config
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session
from cirro_api_client.v1.models import AWSCredentials
from tqdm import tqdm

from cirro.utils import convert_size


def format_creds_for_session(creds: AWSCredentials):
    return {
        'access_key': creds.access_key_id,
        'secret_key': creds.secret_access_key,
        'token': creds.session_token,
        'expiry_time': creds.expiration.isoformat()
    }


class ProgressPercentage:
    def __init__(self, progress: tqdm):
        self._lock = threading.Lock()
        self.progress = progress

    def __call__(self, bytes_amount):
        with self._lock:
            self.progress.update(bytes_amount)


class S3Client:
    def __init__(self, creds_getter: Callable[[], AWSCredentials], enable_additional_checksum=False):
        self._creds_getter = creds_getter
        self._client = self._build_session_client()
        self._upload_args = dict(ChecksumAlgorithm='SHA256') if enable_additional_checksum else dict()
        self._download_args = dict(ChecksumMode='ENABLED') if enable_additional_checksum else dict()

    def upload_file(self, file_path: Path, bucket: str, key: str):
        file_size = file_path.stat().st_size
        file_name = file_path.name

        with tqdm(total=file_size,
                  desc=f'Uploading file {file_name} ({convert_size(file_size)})',
                  bar_format="{desc} | {percentage:.1f}%|{bar:25} | {rate_fmt}",
                  unit='B', unit_scale=True,
                  unit_divisor=1024) as progress:
            with file_path.open('rb') as file:
                self._client.upload_fileobj(file, bucket, key,
                                            Callback=ProgressPercentage(progress),
                                            ExtraArgs=self._upload_args)

    def download_file(self, local_path: Path, bucket: str, key: str):
        file_size = self.get_file_stats(bucket, key)['ContentLength']
        file_name = local_path.name

        with tqdm(total=file_size,
                  desc=f'Downloading file {file_name} ({convert_size(file_size)})',
                  bar_format="{desc} | {percentage:.1f}%|{bar:25} | {rate_fmt}",
                  unit='B', unit_scale=True,
                  unit_divisor=1024) as progress:
            absolute_path = str(local_path.absolute())
            self._client.download_file(bucket, key, absolute_path,
                                       Callback=ProgressPercentage(progress),
                                       ExtraArgs=self._download_args)

    def create_object(self, bucket: str, key: str, contents: str, content_type: str):
        self._client.put_object(
            Bucket=bucket,
            Key=key,
            ContentType=content_type,
            ContentEncoding='utf-8',
            Body=bytes(contents, 'UTF-8'),
            **self._upload_args
        )

    def get_file(self, bucket: str, key: str) -> bytes:
        resp = self._client.get_object(Bucket=bucket, Key=key, **self._download_args)
        file_body = resp['Body']
        return file_body.read()

    def get_file_stats(self, bucket: str, key: str):
        """
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/head_object.html
        """
        return self._client.head_object(Bucket=bucket, Key=key)

    def _build_session_client(self):
        creds = self._creds_getter()

        if creds.expiration:
            session = get_session()
            session._credentials = RefreshableCredentials.create_from_metadata(
                metadata=format_creds_for_session(creds),
                refresh_using=self._refresh_credentials,
                method='sts'
            )
            session = Session(botocore_session=session)
        else:
            session = Session(
                aws_access_key_id=creds.access_key_id,
                aws_secret_access_key=creds.secret_access_key,
                aws_session_token=creds.session_token
            )
        s3_config = Config(
            use_dualstack_endpoint=True
        )
        return session.client('s3', region_name=creds.region, config=s3_config)

    def _refresh_credentials(self):
        new_creds = self._creds_getter()
        return format_creds_for_session(new_creds)



from fnmatch import fnmatch
from pathlib import Path
from typing import List

from cirro_api_client.v1.models import Dataset, Project

from cirro.cli.interactive.common_args import ask_project
from cirro.cli.interactive.utils import ask, prompt_wrapper, InputError
from cirro.cli.models import DownloadArguments
from cirro.models.file import File
from cirro.utils import format_date


def ask_dataset(datasets: List[Dataset], input_value: str) -> str:
    if len(datasets) == 0:
        raise InputError("No datasets available")
    sorted_datasets = sorted(datasets, key=lambda d: d.created_at, reverse=True)
    dataset_prompt = {
        'type': 'autocomplete',
        'name': 'dataset',
        'message': 'What dataset would you like to download? (Press Tab to see all options)',
        'choices': [f'{dataset.name} - {dataset.id}' for dataset in sorted_datasets],
        'meta_information': {
            f'{dataset.name} - {dataset.id}': f'{format_date(dataset.created_at)}'
            for dataset in datasets
        },
        'ignore_case': True
    }
    answers = prompt_wrapper(dataset_prompt)
    choice = answers['dataset']
    # Map the answer to a dataset
    for dataset in datasets:
        if f'{dataset.name} - {dataset.id}' == choice:
            return dataset.id

    # The user has made a selection which does not match
    # any of the options available.
    # This is most likely because there was a typo
    if ask(
        'confirm',
        'The selection does match an option available - try again?'
    ):
        return ask_dataset(datasets, input_value)
    raise InputError("Exiting - no dataset selected")


def ask_dataset_files(files: List[File]) -> List[File]:
    """Get the list of files which the user would like to download from the dataset."""

    choices = [
        "Download all files",
        "Select files from a list",
        "Select files with a naming pattern (glob)"
    ]

    selection_mode_prompt = {
        'type': 'select',
        'name': 'mode',
        'message': 'Which files would you like to download from this dataset?',
        'choices': choices
    }

    answers = prompt_wrapper(selection_mode_prompt)

    if answers['mode'] == choices[0]:
        return files
    elif answers['mode'] == choices[1]:
        return ask_dataset_files_list(files)
    else:
        return ask_dataset_files_glob(files)


def strip_prefix(fp: str, prefix: str):
    assert fp.startswith(prefix), f"Expected {fp} to start with {prefix}"
    return fp[len(prefix):]


def ask_dataset_files_list(files: List[File]) -> List[File]:
    answers = prompt_wrapper({
        'type': 'checkbox',
        'name': 'files',
        'message': 'Select the files to download',
        'choices': [
            strip_prefix(file.relative_path, "data/")
            for file in files
        ]
    })

    selected_files = [
        file
        for file in files
        if strip_prefix(file.relative_path, "data/") in set(answers['files'])
    ]

    if len(selected_files) == 0:
        if ask(
            "confirm",
            "No files were selected - try again?"
        ):
            return ask_dataset_files_list(files)
        else:
            raise InputError("No files selected")
    else:
        return selected_files


def ask_dataset_files_glob(files: List[File]) -> List[File]:

    confirmed = False
    while not confirmed:
        selected_files = ask_dataset_files_glob_single(files)
        confirmed = ask(
            "confirm",
            f'Number of files selected: {len(selected_files):} / {len(files):,}'
        )

    if len(selected_files) == 0:
        raise InputError("No files selected")

    return selected_files


def ask_dataset_files_glob_single(files: List[File]) -> List[File]:

    print("All Files:")
    for file in files:
        print(f" - {strip_prefix(file.relative_path, 'data/')}")

    answers = prompt_wrapper({
        'type': 'text',
        'name': 'glob',
        'message': 'Select files by naming pattern (using the * wildcard)',
        'default': '*'
    })

    selected_files = [
        file
        for file in files
        if fnmatch(strip_prefix(file.relative_path, "data/"), answers['glob'])
    ]

    print("Selected Files:")
    for file in selected_files:
        print(f" - {strip_prefix(file.relative_path, 'data/')}")

    return selected_files


def ask_directory(input_value: str) -> str:
    directory_prompt = {
        'type': 'path',
        'name': 'directory',
        'only_directories': True,
        'message': 'Where would you like to download these files?',
        'default': input_value or str(Path.cwd())
    }

    answers = prompt_wrapper(directory_prompt)
    return answers['directory']


def gather_download_arguments(input_params: DownloadArguments, projects: List[Project]):
    input_params['project'] = ask_project(projects, input_params.get('project'))
    return input_params


def gather_download_arguments_dataset(input_params: DownloadArguments, datasets: List[Dataset]):
    input_params['dataset'] = ask_dataset(datasets, input_params.get('dataset'))
    input_params['data_directory'] = ask_directory(input_params.get('data_directory'))
    return input_params



from typing import List

from cirro_api_client.v1.api.compute_environment import get_compute_environments
from cirro_api_client.v1.models import ComputeEnvironmentConfiguration

from cirro.services.base import BaseService


class ComputeEnvironmentService(BaseService):
    """
    Service for interacting with the Compute Environment endpoints
    """

    def list_environments_for_project(self, project_id: str) -> List[ComputeEnvironmentConfiguration]:
        """
        List of custom compute environments for a project (i.e., an agent)

        Args:
            project_id (str): Project ID

        Returns:
            List of compute environments that are available for the project
        """
        return get_compute_environments.sync(project_id=project_id,
                                             client=self._api_client)



from typing import TypedDict, List, Optional


class DownloadArguments(TypedDict):
    project: str
    dataset: str
    data_directory: str
    interactive: bool


class UploadArguments(TypedDict):
    name: str
    description: str
    project: str
    process: str
    data_directory: str
    include_hidden: bool
    interactive: bool
    files: Optional[List[str]]


class ListArguments(TypedDict):
    project: str
    interactive: bool



import click

from cirro.cli import run_ingest, run_download, run_configure, run_list_datasets
from cirro.cli.controller import handle_error
from cirro.cli.interactive.utils import InputError


def check_required_args(args):
    if args.get('interactive'):
        return
    if any(value is None for value in args.values()):
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit()


@click.group(help="Cirro CLI - Tool for interacting with datasets")
@click.version_option()
def run():
    pass  # Print out help text, nothing to do


@run.command(help='List datasets', no_args_is_help=True)
@click.option('--project',
              help='Name or ID of the project')
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
def list_datasets(**kwargs):
    check_required_args(kwargs)
    run_list_datasets(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Download dataset files', no_args_is_help=True)
@click.option('--project',
              help='Name or ID of the project')
@click.option('--dataset',
              help='ID of the dataset')
@click.option('--file',
              help='Name and relative path of the file (optional)',
              default=[],
              multiple=True)
@click.option('--data-directory',
              help='Directory to store the files')
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
def download(**kwargs):
    check_required_args(kwargs)
    run_download(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Upload and create a dataset', no_args_is_help=True)
@click.option('--name',
              help='Name of the dataset')
@click.option('--description',
              help='Description of the dataset (optional)',
              default='')
@click.option('--project',
              help='Name or ID of the project')
@click.option('--process',
              help='Name or ID of the ingest process')
@click.option('--data-directory',
              help='Directory you wish to upload')
@click.option('-i', '--interactive',
              help='Gather arguments interactively',
              is_flag=True, default=False)
@click.option('--include-hidden',
              help='Include hidden files in the upload (e.g., files starting with .)',
              is_flag=True, default=False)
def upload(**kwargs):
    check_required_args(kwargs)
    run_ingest(kwargs, interactive=kwargs.get('interactive'))


@run.command(help='Configure authentication')
def configure():
    run_configure()


def main():
    try:
        run()
    except InputError as e:
        handle_error(e)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()
